"""
Functions that process data for plotting
"""

import math
from typing import List, Union, Optional, Tuple, Dict

from obspy import UTCDateTime

from sohstationviewer.conf import constants as const
from sohstationviewer.view.util.enums import LogType

maxInt = 1E100
maxFloat = 1.0E100

# TODO: put this in DB
MassPosVoltRanges = {"regular": [0.5, 2.0, 4.0, 7.0],
                     "trillium": [0.5, 1.8, 2.4, 3.5]}
MassPosColorPallets = {"B": ["C", "G", "Y", "R", "M"],
                       "W": ["B", "B", "B", "B", "B"]}


def getMassposValueColors(rangeOpt: str, chan_id: str, cMode: str,
                          processing_log: List[Tuple[str, LogType]],
                          retType: str = 'str'
                          ) -> Optional[Union[str, List[Tuple[float, str]]]]:
    """
    Create a map between value and color based on given rangeOpt and cMode to
    display mass position plots.
    :param rangeOpt: massPosVoltRangeOpt got from Options Menu - MP coloring
        in Main Window to define different values for mass position.
        (regular/trillium)
    :param chan_id: ID of the channel
    :param cMode: color mode (B/W)
    :param processing_log: list of processing info and type
    :param retType: request return type
    :return: [(value, color), (value, color) ...]
        if retType is 'str', return "value:color|value:color"
    """

    if rangeOpt.lower() not in MassPosVoltRanges.keys():
        processing_log.append(
            (
                f"{chan_id}: The current selected Mass Position color range is"
                f" '{rangeOpt}' isn't allowable. The accept ranges are: "
                f"{', '.join(MassPosVoltRanges.keys())}",
                LogType.ERROR
            )
        )
        return
    massPosVoltRange = MassPosVoltRanges[rangeOpt]
    massPosColorPallet = MassPosColorPallets[cMode]
    valueColors = []
    for i in range(len(massPosVoltRange)):
        if retType == 'str':
            valueColors.append(
                "%s:%s" % (massPosVoltRange[i], massPosColorPallet[i]))
        else:
            valueColors.append((massPosVoltRange[i], massPosColorPallet[i]))
        if i == len(massPosVoltRange) - 1:
            if retType == 'str':
                valueColors.append(
                    "%s:+%s" % (
                        massPosVoltRange[i], massPosColorPallet[i + 1]))
            else:
                valueColors.append(
                    (massPosVoltRange[i], massPosColorPallet[i + 1]))
    if retType == 'str':
        return '|'.join(valueColors)
    return valueColors


def formatTime(time: Union[UTCDateTime, float], dateMode: str,
               timeMode: Optional[str] = None) -> str:
    """
    Format time according to date_mode and time_mode
    :param time: time to be format, can be UTCDateTime or epoch time
    :param dateMode: the format of date
    :param timeMode: the format of time
    :return: the formatted time string
    """
    if isinstance(time, UTCDateTime):
        t = time
    else:
        t = UTCDateTime(time)

    # https://docs.python.org/3/library/datetime.html#
    # strftime-and-strptime-format-codes
    format = ''
    if dateMode == 'YYYY-MM-DD':
        format = '%Y-%m-%d'
    elif dateMode == 'YYYYMMDD':
        format = '%Y%m%d'
    elif dateMode == 'YYYY:DOY':
        format = '%Y:%j'
    if timeMode == 'HH:MM:SS':
        format += " %H:%M:%S"

    ret = t.strftime(format)
    return ret


def getTitle(key: Union[str, Tuple[str, int]], minTime: float, maxTime: float,
             dateMode: str) -> str:
    """
    Create title for the plot.
    :param key: str or (str, str) sta_id for mseed, (unit_id, exp_no) for rt130
    :param minTime: start time of the plot
    :param maxTime: end time of the plot
    :param dateMode: format of date
    :return: title for the plot
    """
    diff = maxTime - minTime
    hours = diff / 3600
    return ("%s  %s  to  %s  (%s)" %
            (key,
             formatTime(minTime, dateMode, "HH:MM:SS"),
             formatTime(maxTime, dateMode, "HH:MM:SS"),
             round(hours, 2))
            )


def getGaps(gaps: List[List[float]], gapMin: float
            ) -> List[List[float]]:
    """
    :param gaps: list of gaps
    :param gapMin: minimum of gaps count in minutes
    :return: list of gaps of which gaps smaller than gapMin have been removed
    """
    gapMinSec = gapMin * 60
    return [g for g in gaps if (g[1] - g[0]) >= gapMinSec]


def getTimeTicks(earliest: float, latest: float, dateFormat: str,
                 labelTotal: int
                 ) -> Tuple[List[float], List[float], List[str]]:
    """
    split time range to use for tick labels
    Ex: getTimeTicks(1595542860.0, 1595607663.91, 'YYYY-MM-DD', 3)
    :param earliest: earliest epoch time
    :param latest: latest epoch time
    :param dateFormat: (YYYY:DOY, YYYY-MM-DD or YYYYMMMDD)
        (selected in Menu Options - Date Format)
    :param labelTotal: number of time label to be displayed,
        others will show as ticks oly
    :return:
        times: list of times to show ticks
        majorTimes: list of times for displayed labels
        majorTimelabels: list of labels displayed
    """
    timeRange = latest - earliest
    if timeRange >= 2592000.0:
        mode = "DD"
        # Time of the nearest midnight before the earliest time.
        time = (earliest // 86400.0) * 86400
        interval = 864000.0
    elif timeRange >= 864000.0:
        mode = "D"
        time = (earliest // 86400.0) * 86400
        interval = 86400.0
    elif timeRange >= 3600.0:
        mode = "H"
        # Nearest hour.
        time = (earliest // 3600.0) * 3600
        interval = 3600.0
    elif timeRange >= 60.0:
        mode = "M"
        # Nearest minute.
        time = (earliest // 60.0) * 60
        interval = 60.0
    else:
        mode = "S"
        time = (earliest // 1)
        interval = 1.0
    times = []
    timeLabels = []
    time += interval
    while time < latest:
        times.append(time)
        timeLabel = formatTime(time, dateFormat, 'HH:MM:SS')
        if mode == "DD" or mode == "D":
            timeLabel = timeLabel[:-9]
        elif mode == "H":
            timeLabel = timeLabel[:-3]
        elif mode == "M" or mode == "S":
            timeLabel = timeLabel
        timeLabels.append(timeLabel)
        time += interval

    ln = len(timeLabels)
    d = math.ceil(len(timeLabels) / labelTotal)
    majorTimes = [times[i] for i in range(ln) if i % d == 0]
    majorTimelabels = [timeLabels[i] for i in range(ln) if i % d == 0]
    return times, majorTimes, majorTimelabels


def getDayTicks() -> Tuple[List[int], List[int], List[str]]:
    """
    Get information for displaying time on plotting widget.
    :return:
        times: list of indexes of every hour in each_day_5_min_list
        majorTimes: list of indexes of every 4 hours in each_day_5_min_list
        majorTimeLabels: 2 digit numbers of every 4 hours in a day
    """

    times = list(range(const.NO_5M_1H, const.NO_5M_DAY, const.NO_5M_1H))
    majorTimes = list(range(4 * const.NO_5M_1H,
                            const.NO_5M_DAY,
                            4 * const.NO_5M_1H))
    majorTimeLabels = ["%02d" % int(t / const.NO_5M_1H) for t in majorTimes]
    return times, majorTimes, majorTimeLabels


def getUnitBitweight(chanDB: Dict, bitweightOpt: str) -> str:
    """
    Get the format to diplay value including fixed point decimal and unit based
    on the information from the database.
    :param chanDB: channel's info got from database
    :param bitweightOpt: option for bitweight (none, low, high)
        (Menu Options - Q330 Gain)
    :return unitBitweight: format for displayed value on the left of each
        plot with fixed point decimal and unit
    """

    plotType = chanDB['plotType']
    # Not all channels use/have this field.
    unit = chanDB['unit']
    try:
        fixPoint = chanDB['fixPoint']
    except Exception:
        fixPoint = 0
    unitBitweight = ''

    if plotType in ['linesDots', 'linesSRate', 'linesMasspos']:

        if fixPoint == 0:
            unitBitweight = "{}%s" % unit
        else:
            unitBitweight = "{:.%sf}%s" % (fixPoint, unit)
    if chanDB['channel'] == 'SEISMIC':
        if fixPoint != 0:
            unitBitweight = "{:.%sf}%s" % (fixPoint, unit)
        else:
            if bitweightOpt in ["low", "high"]:
                unitBitweight = "{}V"
            else:
                unitBitweight = "{}%s" % unit
    return unitBitweight
