"""
Function that ignite from main_window, Dialogs to read data files for data,
channels, datatype
"""

import os
import json
import re
import traceback
from pathlib import Path
from typing import List, Set, Optional, Dict, Tuple

from PySide2.QtWidgets import QTextBrowser
from obspy.core import read as read_ms
from obspy.io.reftek.core import Reftek130Exception

from sohstationviewer.model.mseed.mseed import MSeed
from sohstationviewer.database.extract_data import get_signature_channels
from sohstationviewer.model.data_type_model import DataTypeModel

from sohstationviewer.controller.util import validateFile, displayTrackingInfo

from sohstationviewer.view.util.enums import LogType


def loadData(dataType: str, tracking_box: QTextBrowser, listOfDir: List[str],
             reqWFChans: List[str] = [], reqSOHChans: List[str] = [],
             readStart: Optional[float] = None,
             readEnd: Optional[float] = None) -> DataTypeModel:
    """
    Load the data stored in listOfDir and store it in a DataTypeModel object.
    The concrete class of the data object is based on dataType. Run on the same
    thread as its caller, and so will block the GUI if called on the main
    thread. It is advisable to use model.data_loader.DataLoader to load data
    unless it is necessary to load data in the main thread (e.g. if there is
    a need to access the call stack).

    :param dataType: type of data read
    :param tracking_box: widget to display tracking info
    :param listOfDir: list of directories selected by users
    :param reqWFChans: requested waveform channel list
    :param reqSOHChans: requested soh channel list
    :param readStart: start time of read data
    :param readEnd: finish time of read data
    :return dataObject: object that keep the data read from
        listOfDir
    """
    dataObject = None
    for d in listOfDir:
        if dataObject is None:
            try:
                dataObject = DataTypeModel.create_data_object(
                    dataType, tracking_box, d,
                    reqWFChans=reqWFChans, reqSOHChans=reqSOHChans,
                    readStart=readStart, readEnd=readEnd)
            except Exception:
                fmt = traceback.format_exc()
                msg = f"Dir {d} can't be read due to error: {str(fmt)}"
                displayTrackingInfo(tracking_box, msg, LogType.WARNING)

            # if dataObject.hasData():
            #     continue
            # If no data can be read from the first dir, throw exception
            # raise Exception("No data can be read from ", d)
        # TODO: will work with select more than one dir later
        # else:
        #     dataObject.readDir(d)

    # return dataObject.plottingData
    return dataObject


def readChannels(tracking_box: QTextBrowser, listOfDir: List[str]
                 ) -> Set[str]:
    """
    Scan available channels (to be used in channel preferences dialog). Since
        channels for RT130 is hard code, this function won't be applied for it.
    :param tracking_box: widget to display tracking info
    :param listOfDir: list of directories selected by users
    :return dataObject.channels: set of channels present in listofDir
    """
    dataObject = None
    for d in listOfDir:
        if dataObject is None:
            # dataObject = Reftek.Reftek(parent, d)
            # if dataObject.hasData():
            #     continue

            # dataObject = MSeed_Text(parent, d, readChanOnly=True)
            dataObject = MSeed(tracking_box, d, readChanOnly=True)
            if len(dataObject.channels) == 0:
                # If no data can be read from the first dir, throw exception
                raise Exception("No data can be read from ", d)
        else:
            dataObject.readDir(d, readChanOnly=True)
    return dataObject.channels


def detectDataType(tracking_box: QTextBrowser, listOfDir: List[str]
                   ) -> Optional[str]:
    """
    Detect data type for the given directories using getDataTypeFromFile
    :param tracking_box: widget to display tracking info
    :param listOfDir: list of directories selected by users
    :return:
        + if there are more than one data types detected,
            return None with a warning message
        + if only Unknown data type detected,
            return None with a warning message
        + if data type found, return data_type,
    """
    sign_chan_dataType_dict = get_signature_channels()

    dirDataTypeDict = {}
    for d in listOfDir:
        dataType = "Unknown"
        for path, subdirs, files in os.walk(d):
            for fileName in files:
                path2file = Path(path).joinpath(fileName)
                if not validateFile(path2file, fileName):
                    continue
                ret = getDataTypeFromFile(path2file, sign_chan_dataType_dict)
                if ret is not None:
                    dataType, chan = ret
                    break
            if dataType != "Unknown":
                break
        if dataType == "Unknown":
            dirDataTypeDict[d] = ("Unknown", '_')
        else:
            dirDataTypeDict[d] = (dataType, chan)
    dataTypeList = {d[0] for d in dirDataTypeDict.values()}
    if len(dataTypeList) > 1:
        dirDataTypeStr = json.dumps(dirDataTypeDict)
        dirDataTypeStr = re.sub(r'\{|\}|"', '', dirDataTypeStr)
        dirDataTypeStr = re.sub(r', ', '\n', dirDataTypeStr)
        msg = (f"There are more than one types of data detected:\n"
               f"{dirDataTypeStr}\n\n"
               f"Please have only data that related to each other.")
        displayTrackingInfo(tracking_box, msg, LogType.ERROR)
        return

    elif dataTypeList == {'Unknown'}:
        msg = ("There are no known data detected.\n"
               "Please select different folder(s).")
        displayTrackingInfo(tracking_box, msg, LogType.ERROR)
        return

    return list(dirDataTypeDict.values())[0][0]


def getDataTypeFromFile(
        path2file: Path,
        sign_chan_data_type_dict: Dict[str, str]
) -> Optional[Tuple[str, str]]:
    """
    + Try to read mseed data from given file
        if catch TypeError: no data type detected => return None
        if catch Reftek130Exception: data type => return data type RT130
        otherwise data type is mseed which includes: q330, pegasus, centaur
    + Continue to identify data type for a file by checking if the channel
    in that file is a unique channel of a data type.
    :param path2file: absolute path to processed file
    :param sign_chan_data_type_dict: dict of unique chan for data
        type
    :return: detected data type, channel from which data type is detected
    """
    try:
        stream = read_ms(path2file)
    except TypeError:
        return
    except Reftek130Exception:
        return 'RT130', '_'

    for trace in stream:
        chan = trace.stats['channel']
        if chan in sign_chan_data_type_dict.keys():
            return sign_chan_data_type_dict[chan], chan
