"""
basic functions: format, validate, display tracking
"""

import os
import re
from datetime import datetime
from pathlib import Path

from PySide2 import QtCore
from typing import Tuple, Union

from PySide2.QtWidgets import QTextBrowser
from obspy import UTCDateTime
import numpy as np
from sohstationviewer.view.util.enums import LogType


def validateFile(path2file: Union[str, Path], fileName: str):
    """
    Check if fileName given is a file and not info file
    :param path2file: absolute path to file
    :param fileName: name of the file
    :return: True if pass checking, False if not.
    """

    if fileName.strip() == '.DS_Store' or fileName.startswith('._'):
        # skip mac's info file
        return False

    if not os.path.isfile(path2file):
        return False
    return True


@QtCore.Slot()
def displayTrackingInfo(trackingBox: QTextBrowser, text: str,
                        type: LogType = LogType.INFO):
    """
    Display text in the given widget with different background and text colors
    :param trackingBox: widget to display tracking info
    :param text: info to be displayed
    :param type: (info/warning/error) type of info to be displayed in
        different color
    """

    if trackingBox is None:
        print(f"{type.name}: {text}")
        return

    msg = {'text': text}
    if type == LogType.ERROR:
        msg['color'] = 'white'
        msg['bgcolor'] = '#e46269'
    elif type == LogType.WARNING:
        msg['color'] = '#ffd966'
        msg['bgcolor'] = 'orange'
    else:
        msg['color'] = 'blue'
        msg['bgcolor'] = 'white'
    htmlText = """<body>
        <div style='color:%(color)s; background-color:%(bgcolor)s'>
            %(text)s
        </div>
        </body>"""
    trackingBox.setHtml(htmlText % msg)
    # parent.update()
    trackingBox.repaint()


def getDirSize(dir: str) -> Tuple[int, int]:
    """
    Get size of directory and size of file.
    :param dir: absolute path to directory
    :return:
        totalSize: total size of the directory
        totalFile: total file of the directory
    """

    totalSize = 0
    totalFile = 0
    for path, subdirs, files in os.walk(dir):
        for fileName in files:
            if not validateFile(os.path.join(path, fileName), fileName):
                continue
            fp = os.path.join(path, fileName)
            # print("file %s: %s" % (fp, os.path.getsize(fp)))
            totalSize += os.path.getsize(fp)
            totalFile += 1
    return totalSize, totalFile


def getTime6(timeStr: str) -> Tuple[float, int]:
    """
    Get time from 6 parts string.
    (year:day of year:hour:minute:second:millisecond)
    Ex: 01:251:09:41:35:656/ 2001:251:09:41:35:656
    in which year in the first part can be 2 digits or 6 digits
    :param timeStr: 6 part time string
    :return the epoch time and the year of timeStr.
    """
    year = timeStr.split(':')[0]
    if len(year) == 2:
        return getTime6_2y(timeStr)
    else:
        return getTime6_4y(timeStr)


def getTime6_2y(timeStr: str) -> Tuple[float, int]:
    """
    Get time from 6 parts string in which year has 2 digits.
    Ex: 01:251:09:41:35:656
    :param timeStr: 6 part time string with 2 digits for year
    :return the epoch time and the year of timeStr.
    """
    # pad 0 so the last part has 6 digits to match with the format str
    timeStr = timeStr.ljust(22, "0")
    time = datetime.strptime(timeStr, "%y:%j:%H:%M:%S:%f")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year


def getTime6_4y(timeStr: str) -> Tuple[float, int]:
    """
    Get time from 6 parts string in which year has 4 digits.
    Ex: 2001:251:09:41:35:656
    :param timeStr: 6 part time string with 4 digits for year
    :return the epoch time and the year of timeStr.
    """
    # pad 0 so the last part has 6 digits to match with the format str
    timeStr = timeStr.ljust(24, "0")
    time = datetime.strptime(timeStr, "%Y:%j:%H:%M:%S:%f")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year


def getTime4(timeStr: str, trackingYear: int, yAdded: bool
             ) -> Tuple[float, int, bool]:
    """
    Get time from 4 parts string. (day of year:hour:minute:second)
    Ex: 253:19:41:42
    :param timeStr: time string
    :param trackingYear: year that has been detected
    :param yAdded: flag to tell if year has been plussed 1 or not
    :return:
        + utcTime.timestamp: epoch time
        + time.year: year
        + yAdded: flag to tell if year has been plussed 1 or not
    """
    if not yAdded:
        # first day => move to next year
        doy = int(timeStr.split(':')[0])
        if doy == 1:
            trackingYear += 1
            yAdded = True
    timeStr = f'{str(trackingYear)}:{timeStr}'
    time = datetime.strptime(timeStr, "%Y:%j:%H:%M:%S")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year, yAdded


def getVal(text: str) -> float:
    """
    Get the value part of a string with non-number substring following.
    :param text: value string including unit
    :return: value part including +/-, remove str that follows
    """
    REVal = '^\+?\-?[0-9]+\.?[0-9]?'            # noqa: W605
    return float(re.search(REVal, text).group())


def isBinaryStr(text):
    """
    :param text: text to check
    :return: True if text is a binary string or False if not
    """

    return lambda b: bool(b.translate(None, text))


def rtnPattern(text: str, upper: bool = False) -> str:
    """
    This function is from logpeek's rtnPattern.
    return routine pattern of the string with:
     + 0 for digit
     + a for lowercase
     + A for upper case
     + remain special character
    :param text: text to get format
    :param upper: flag of whether to convert all alphabetic characters
        to A
    :return rtn: routine pattern of the string
    """
    rtn = ""
    for c in text:
        if c.isdigit():
            rtn += "0"
        elif c.isupper():
            rtn += "A"
        elif c.islower():
            rtn += "a"
        else:
            rtn += c
    # So the A/a chars will always be A, so the caller knows what to look for
    if upper is True:
        return rtn.upper()
    return rtn


def add_thousand_separator(Value: float) -> str:
    """
    This function is from logpeek's fmti
    Given Value will be convert to a string integer with thousand separators
    :param Value: string of value with unit
    :return NewValue: new value with no unit and with thousand separators
    """
    Value = int(Value)
    if Value > -1000 and Value < 1000:
        return str(Value)
    Value = str(Value)
    NewValue = ""
    # There'll never be a + sign.
    if Value[0] == "-":
        Offset = 1
    else:
        Offset = 0
    CountDigits = 0
    for i in range(len(Value) - 1, -1 + Offset, -1):
        NewValue = Value[i] + NewValue
        CountDigits += 1
        if CountDigits == 3 and i != 0:
            NewValue = "," + NewValue
            CountDigits = 0
    if Offset != 0:
        if NewValue.startswith(","):
            NewValue = NewValue[1:]
        NewValue = Value[0] + NewValue
    return NewValue


def apply_convert_factor(c_data: dict, convert_factor: float):
    """
    convertFactor = 150mV/count = 150V/1000count
    => unit data * convertFactor= data *150/1000 V

    :param c_data: data of the channel which includes down-sampled
        data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
        soh_data[key][chan_id]
    :param convert_factor: convertFactor field retrieved from
        db table Channels for this channel
    """
    c_data['data'] = np.multiply(c_data['data'], [convert_factor])
