from __future__ import annotations
import os

from tempfile import mkdtemp
import shutil
from typing import Optional, Union, List, Tuple, Dict

from PySide2 import QtCore

from sohstationviewer.controller.util import displayTrackingInfo
from sohstationviewer.conf import constants
from sohstationviewer.model.gps_point import GPSPoint
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.database.process_db import execute_db

from PySide2 import QtWidgets


class WrongDataTypeError(Exception):
    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)


class ThreadStopped(Exception):
    """
    An exception that is raised when the user requests for the data loader
    thread to be stopped.
    """
    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)


class DataTypeModel():
    def __init__(self, trackingBox: QtWidgets.QTextBrowser, folder: str,
                 readChanOnly: bool = False,
                 reqWFChans: Union[List[str], List[int]] = [],
                 reqSOHChans: List[str] = [], readStart: float = 0,
                 readEnd: float = constants.HIGHEST_INT,
                 creator_thread: Optional[QtCore.QThread] = None,
                 notification_signal: Optional[QtCore.Signal] = None,
                 pause_signal: Optional[QtCore.Signal] = None,
                 *args, **kwargs):
        """
        Super class for different data type to process data from data files

        :param trackingBox: widget to display tracking info
        :param folder: path to the folder of data
        :param readChanOnly: if True, only read for channel name
        :param reqWFChans: requested waveform channel list
        :param reqSOHChans: requested SOH channel list
        :param readStart: requested start time to read
        :param readEnd: requested end time to read
        :param creator_thread: the thread the current DataTypeModel instance is
            being created in. If None, the DataTypeModel instance is being
            created in the main thread
        :param notification_signal: signal used to send notifications to the
            main thread. Only not None when creator_thread is not None
        :param pause_signal: signal used to notify the main thread that the
            data loader is paused.
        """
        self.trackingBox = trackingBox
        self.dir = folder
        self.reqSOHChans = reqSOHChans
        self.reqWFChans = reqWFChans
        self.readChanOnly = readChanOnly
        self.readStart = readStart
        self.readEnd = readEnd
        if creator_thread is None:
            err_msg = (
                'A signal is not None while running in main thread'
            )
            assert notification_signal is None, err_msg
            assert pause_signal is None, err_msg
            self.creator_thread = QtCore.QThread()
        else:
            self.creator_thread = creator_thread
        self.notification_signal = notification_signal
        self.pause_signal = pause_signal
        """
        processingLog: [(message, type)] - record the progress of processing
        """
        self.processingLog: List[Tuple[str, LogType]] = []

        DataKey = Union[Tuple[str, str], str]

        """
        Log data: info from log channels, soh messages, text file in dict:
        {'TEXT': [str,], key:{chan_id: [str,],},}
        In which 'TEXT': is the chan_id given by sohview for text only file.
        Note: logData for RT130's dataset has only one channel: SOH
        """
        self.logData = {'TEXT': []}  # noqa

        """
        waveformData: data of waveform in dict:
        {set_key: {
            'files_info': {
                chan_id - str: [{
                    'path2file': absolute path to file - str,
                    'file_name': name of file - str,
                    'nets': list of network names in file - list of str
                    'stats': list of station names in file - list of str
                    'net_stats': list of (network, station)s in file - list of
                        tuple of str
                    'chan_ids': list of channel names in file - list of str
                    'start_epoch': start time of file - float
                    'end_epoch': end time of file - float
                    'traces_total': length of data - int
                    'spr': sample rate - float
                    'read': file has been read or not - bool
                    }]
            }
            'read_data': {
                chan_id - str: {
                    'samplerate': Sample rate of the data - float
                    'tracesInfo': [{
                        'samplerate': Sample rate of the data - float
                        'start_tm_epoch': Start epoch time of the trace - float
                        'end_tm_epoch': End epoch time of the trace - float
                        'size': Size of data - int
                        'times_f': np.memmap's file path for real epoch time
                            - str
                        'data_f': np.memmap's file path for data - str
                        },]
                    'times': times that has been trimmed and down- sampled for
                        plotting - np.array of float,
                    'data': data that has been trimmed and down-sampled for
                        plotting - np.array of float/int,
                    'chan_db_info': the plotting parameters got from database
                        for this channel - dict,
                    ax_wf: axes to draw the channel in WaveformWidget, to
                        combine 2 channels in need - matplotlib.axes.Axes
                }
            }
        }
        (np.memmap is used instead of data for waveform because waveform data
        are too large and make system crashed when try to merge and process)
        """
        self.waveformData = {}

        """
        SOHData: data for SOH info in dict:
        {set_key - str or (str, str): {
            chan_id - str: {
                'org_trace': {
                    'chan_id': channel name - str
                    'samplerate': sample rate of the data - float
                    'start_tm_epoch': start epoch time of the trace - float,
                    'end_tm_epoch': end epoch time of the trace - float,
                    'times': data's real time in epoch - np.array of float,
                    'data': data - np.array of float,
                    'logIdx: soh message line indexes - np.array of int,
                    }
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
                logIdx: soh message line indexes that has been trimmed and
                    down-sampled for plotting - np.array of int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.SOHData = {}

        """
        massPosData: data for mass position info in the format similar to
        SOHData but there are 2 axes for drawing a mass possition channel,
        one is 'ax' to display in PlottingWidget, the other is 'ax_wf' to
        display in WaveformWidget.
        Channel name for mseed mass position: VP*
        Channel name for reftek mass position: MP* (given by SOHView)
                {set_key - str or (str, str): {
            chan_id - str: {
                'org_trace': {
                    'chan_id': channel name - str
                    'samplerate': sample rate of the data - float
                    'start_tm_epoch': start epoch time of the trace - float,
                    'end_tm_epoch': end epoch time of the trace - float,
                    'times': data's real time in epoch - np.array of float,
                    'data': data - np.array of float,
                    }
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes,
                ax_wf: axes to draw the channel in WaveformWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.massPosData = {}

        """
        dataTime: time range of data sets:
        {setKey: [earliestepoch, latestepoch]} - {str: [float, float],}
        """
        self.dataTime: Dict[DataKey, List[float]] = {}

        """
        The given data may include more than one data set which is station_id
        in mseed or (unit_id, exp_no) in reftek. User are allow to choose which
        data set to be displayed
        selectedKey: str - key of the data set to be displayed
        """
        self.selectedKey: Optional[str] = None

        """
        gaps: gaps info in dict:
        {set_key: [list of gaps]} - {str: [[float, float],],}
        """
        self.gaps: Dict[DataKey, List[List[float]]] = {}

        """
         tmpDir: str - dir to keep memmap files.
         Will be deleted when object is deleted
        """
        self.tmpDir: str = mkdtemp()
        self.save_temp_data_folder_to_database()
        try:
            os.mkdir(self.tmpDir)
        except FileExistsError:
            shutil.rmtree(self.tmpDir)
            os.mkdir(self.tmpDir)

        self._pauser = QtCore.QSemaphore()
        self.pause_response = None

        self.gps_points: List[GPSPoint] = []

    def __del__(self):
        print("delete dataType Object")
        try:
            shutil.rmtree(self.tmpDir)
        except OSError as e:
            self.trackInfo(
                "Error deleting %s : %s" % (self.tmpDir, e.strerror),
                LogType.ERROR)
            print("Error deleting %s : %s" % (self.tmpDir, e.strerror))
        print("finish deleting")

    def hasData(self) -> bool:
        """
        :return bool - True if there is any data can be read.
                       False if there is no valid data
        """
        if (len(self.logData) == 0 and len(self.SOHData) == 0 and
                len(self.massPosData) == 0 and len(self.waveformData) == 0):
            return False
        return True

    def trackInfo(self, text: str, type: LogType) -> None:
        """
        Display tracking info in tracking_box.
        Add all errors/warnings to processing_log.
        :param text: str - message to display
        :param type: str - type of message (error/warning/info)
        """
        # displayTrackingInfo updates a QtWidget, which can only be done in the
        # main thread. So, if we are running in a background thread
        # (i.e. self.creator_thread is not None), we need to use signal slot
        # mechanism to ensure that displayTrackingInfo is run in the main
        # thread.
        if self.notification_signal is None:
            displayTrackingInfo(self.trackingBox, text, type)
        else:
            self.notification_signal.emit(self.trackingBox, text, type)
        if type != LogType.INFO:
            self.processingLog.append((text, type))

    @classmethod
    def create_data_object(cls, data_type, tracking_box, folder,
                           readChanOnly=False, reqWFChans=[], reqSOHChans=[],
                           readStart=0, readEnd=constants.HIGHEST_INT):
        """
        Create a DataTypeModel object, with the concrete class being based on
        data_type. Run on the same thread as its caller, and so will block the
        GUI if called on the main thread. Do not call this method directly.
        Instead, call the wrapper controller.processing.loadData.

        :param data_type: str - type of data read
        :param tracking_box: QTextBrowser - widget to display tracking info
        :param folder: [str,] - the data directory
        :param readChanOnly: if True, only read channel name
        :param reqWFChans: [str,] - requested waveform channel list
        :param reqSOHChans: [str,] - requested soh channel list
        :param readStart: [float,] - start time of read data
        :param readEnd: [float,] - finish time of read data
        :return: DataTypeModel - object that keep the data read from
            folder
        """
        if data_type == 'RT130':
            from sohstationviewer.model.reftek.reftek import RT130
            dataObject = RT130(
                tracking_box, folder, readChanOnly=readChanOnly,
                reqWFChans=reqWFChans, reqSOHChans=reqSOHChans,
                readStart=readStart, readEnd=readEnd)
        else:
            from sohstationviewer.model.mseed.mseed import MSeed
            dataObject = MSeed(
                tracking_box, folder, readChanOnly=readChanOnly,
                reqWFChans=reqWFChans, reqSOHChans=reqSOHChans,
                readStart=readStart, readEnd=readEnd)
        return dataObject

    def pause(self) -> None:
        """
        Pause the thread this DataTypeModel instance is in. Works by trying
        to acquire a semaphore that is not available, which causes the thread
        to block.

        Note: due to how this is implemented, each call to pause will require
        a corresponding call to unpause. Thus, it is inadvisable to call this
        method more than once.

        Caution: not safe to call in the main thread. Unless a background
        thread releases the semaphore, the whole program will freeze.
        """
        self._pauser.acquire()

    @QtCore.Slot()
    def unpause(self):
        """
        Unpause the thread this DataTypeModel instance is in. Works by trying
        to acquire a semaphore that is not available, which causes the thread
        to block.

        Caution: due to how this is implemented, if unpause is called before
        pause, the thread will not be paused until another call to pause is
        made. Also, like pause, each call to unpause must be matched by another
        call to pause for everything to work.
        """
        self._pauser.release()

    @QtCore.Slot()
    def receive_pause_response(self, response: object):
        """
        Receive a response to a request made to another thread and unpause the
        calling thread.

        :param response: the response to the request made
        :type response: object
        """
        self.pause_response = response
        self.unpause()

    @classmethod
    def get_empty_instance(cls) -> DataTypeModel:
        """
        Create an empty data object. Useful if a DataTypeModel instance is
        needed, but it is undesirable to load a data set. Basically wraps
        __new__().

        :return: an empty data object
        :rtype: DataTypeModel
        """
        return cls.__new__(cls)

    def save_temp_data_folder_to_database(self):
        execute_db(f'UPDATE PersistentData SET FieldValue="{self.tmpDir}" '
                   f'WHERE FieldName="tempDataDirectory"')
