"""
Functions that help processing model data
"""
import math
from struct import unpack
from pathlib import Path
from typing import Dict, Callable, Tuple, List, Union, IO, Optional

import numpy as np
from obspy.core import Trace, Stream, read as read_ms
from PySide2 import QtWidgets
from obspy.io.reftek.core import Reftek130

from sohstationviewer.model.mseed.blockettes_reader import (
    readNextBlkt, ReadBlocketteError)
from sohstationviewer.conf.dbSettings import dbConf
from sohstationviewer.conf import constants as const
from sohstationviewer.model.reftek.from_rt2ms import core
from sohstationviewer.view.util.enums import LogType


def readSOHMSeed(path2file: Path, fileName: str,
                 SOHStreams: Dict[str, Dict[str, Stream]],
                 logData: Dict[str, List[str]],
                 netsProbInFile: Dict[Tuple[str, ...], str],
                 trackInfo: Callable[[str, LogType], None]) -> Dict[str, List]:
    """
    Use read() from obspy.core to read miniseed file:
       + if encoding is ASCII: log string will be added to log_data. The
           following blockettes' data if available will be converted to string
           to add to log_data as well.
       + otherwise traces from stream will be add to soh_streams to be
           merged later
    :param path2file: absolute path of mseed file
    :param fileName: name of mseed file
    :param SOHStreams: holder of SOH mseed streams
    :param logData: holder of info from log
    :param netsProbInFile: holder of dict with key are
        nets of file and value is user-selected net so the rule will be
        applied for other files if their nets is subset of a key
    :param trackInfo: to display processing info
    """
    stream = read_ms(path2file)

    file = None
    nets = set()
    stats = set()
    netStats = set()
    chanIDs = set()
    newnetID = None
    for trace in stream:
        netID = trace.stats['network'].strip()
        if netID != "None":
            nets.add(netID)
        if len(nets) > 1:
            for netProb in netsProbInFile:
                netProbSet = set(netProb)
                if nets.issubset(netProbSet) or netProbSet.issubset(nets):
                    newnetID = netsProbInFile[netProb]
                    if nets != netProbSet:
                        netsProbInFile[tuple(nets)] = newnetID
                    break
            if newnetID is None:
                msg = (f"There are more than one net in file {fileName}.\n"
                       "Please select one.")
                msgBox = QtWidgets.QMessageBox()
                msgBox.setText(msg)
                netButtons = []
                _nets = sorted(list(nets))
                for net in _nets:
                    netButtons.append(msgBox.addButton(
                        net, QtWidgets.QMessageBox.ActionRole))
                msgBox.exec_()

                selectedIdx = netButtons.index(msgBox.clickedButton())
                newnetID = _nets[selectedIdx]
                netsProbInFile[tuple(nets)] = newnetID

    for trace in stream:
        if newnetID is not None:
            # use one net ID for all traces in a file
            trace.stats['network'] = newnetID
        netID = trace.stats['network'].strip()
        chanID = trace.stats['channel'].strip()
        staID = trace.stats['station'].strip()
        stats.add(staID)
        chanIDs.add(chanID)
        netStats.add((netID, staID))
        if trace.stats.mseed['encoding'] == 'ASCII':
            file = readASCII(
                path2file, file, staID, chanID, trace, logData, trackInfo)
        else:
            if staID not in SOHStreams:
                SOHStreams[staID] = {}
            if chanID not in SOHStreams[staID]:
                SOHStreams[staID][chanID] = Stream()
            SOHStreams[staID][chanID].append(trace)

    if file is not None:
        file.close()
    return {'nets': list(nets),
            'stats': sorted(list(stats)),
            'netStats': sorted(list(netStats)),
            'chanIDs': sorted(list(chanIDs))
            }


def readSOHTrace(trace: Trace) -> Dict:
    """
    Read SOH trace's info
    :param trace: mseed trace
    :return tr: with trace's info
        (structure in DataTypeModel.__init__.soh_ata[key][chan_id][orgTrace])
    """
    tr = {}
    tr['chanID'] = trace.stats['channel']
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    tr['times'] = trace.times() + trace.stats['starttime'].timestamp
    tr['data'] = trace.data
    return tr


def readMPTrace(trace: Trace) -> Dict:
    """
    Read mass possition trace's info using readSOHTrace(), then calculate real
        value for mass possition
    :param trace: mseed trace
    :return tr: with trace's info from readSOHTrace in which tr['data']
        has been converted from 16-bit signed integer in which
        32767= 2 ** 16/2 - 1 is the highest value of 16-bit two's complement
        number. The value is also multiplied by 10 for readable display.
        (structure in DataTypeModel.__init__.soh_data[key][chan_id][orgTrace])
    (According to 130_theory.pdf: Each channel connects to a 12-bit A/D
    converter with an input range of +/- 10V. These channel are read
    once per second as left-justified, 2's-compliment, 16 bit values.)

    """
    tr = readSOHTrace(trace)
    tr['data'] = np.round_(tr['data'] / 32767.0 * 10.0, 1)
    return tr


def readWaveformTrace(trace: Trace, sta_id: Union[Tuple[str, str], str],
                      chan_id: str, traces_info: List, tmp_dir: str) -> Dict:
    """
    read mseed waveform trace and save data to files to save mem for processing
        since waveform data are big.
    :param trace: mseed trace
    :param sta_id: station name
    :param chan_id: channel name
    :param traces_info: holder of traces_info, refer
        DataTypeModel.__init__.
        waveform_data[key]['read_data'][chan_id]['traces_info']
    :param tmp_dir: path to the directory that store memmap files
    :return tr: with trace's info
        (structure in DataTypeModel.__init__.
         waveform_data[key][chan_id][traces_info])
    """
    # gaps for SOH only for now
    tr = {}
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    times = trace.times() + trace.stats['starttime'].timestamp
    data = trace.data
    tr['size'] = times.size

    trIdx = len(traces_info)
    tr['times_f'] = saveData2File(
        tmp_dir, 'times', sta_id, chan_id, times, trIdx, tr['size'])
    tr['data_f'] = saveData2File(
        tmp_dir, 'data', sta_id, chan_id, data, trIdx, tr['size'])
    return tr


def readWaveformMSeed(path2file: str, fileName: str, staID: str, chanID: str,
                      tracesInfo: List, dataTime: List[float], tmpDir: str
                      ) -> None:
    """
    Read traces from waveform mseed file to append to tracesInfo.
        data_time is update for new min and max time.
    :param path2file: absolute path to waveform mseed file
    :param fileName: name of waveform mseed file
    :param staID: station ID from indexing
    :param chanID: channel ID from indexing
    :param tracesInfo: holder of traces_info, refer
        DataTypeModel.__init__.
        waveform_data[key]['read_data'][chan_id]['traces_info']
    :param dataTime: DataTypeModel.__init__.data_time[key] - holder for data
        time of the current station
    :param tmpDir: the folder to keep memmap files
    """
    stream = read_ms(path2file)
    for trace in stream:
        tr = readWaveformTrace(trace, staID, chanID, tracesInfo, tmpDir)
        dataTime[0] = min(tr['startTmEpoch'], dataTime[0])
        dataTime[1] = max(tr['endTmEpoch'], dataTime[1])
        tracesInfo.append(tr)


def readWaveformReftek(rt130: Reftek130, key: Tuple[str, str], read_data: Dict,
                       dataTime: List[float], tmpDir: str) -> None:
    """
    Read traces from rt130 object to add to traces Info tracesInfo.
    dataTime is update for new min and max time.
    :param rt130: rt130 object
    :param key: (str, str) - key of the data set
    :param read_data: dict -holder of the readData refer
        DataTypeModel.__init__.waveform_data[key]['read_data']
    :param dataTime: DataTypeModel.__init__.data_time[key] - holder for data
        time of the current station
    :param tmpDir: str - the folder to keep memmap files
    """
    stream = core.Reftek130.to_stream(
        rt130,
        headonly=False,
        verbose=False,
        sort_permuted_package_sequence=True)
    for trace in stream:
        chanID = trace.stats['channel']
        samplerate = trace.stats['sampling_rate']
        if chanID not in read_data:
            read_data[chanID] = {
                "tracesInfo": [],
                "samplerate": samplerate}
        tracesInfo = read_data[chanID]['tracesInfo']
        tr = readWaveformTrace(trace, key, chanID, tracesInfo, tmpDir)
        dataTime[0] = min(tr['startTmEpoch'], dataTime[0])
        dataTime[1] = max(tr['endTmEpoch'], dataTime[1])
        tracesInfo.append(tr)


def readASCII(path2file: Path, file: IO, sta_id: str, chan_id: str,
              trace: Trace, log_data: dict,
              track_info: Callable[[str, LogType], None]) -> IO:
    """
    Read mseed trace with ASCII encoding to add to logData.
    :param path2file: str- absolute path of mseed file
    :param file: file object - to continue reading. Open new if file is None
    :param sta_id: str - station ID got from mseed header
    :param chan_id: str - channel ID got from mseed header
    :param trace: obspy.core.trace.Trace - mseed trace
    :param log_data: dict - holder for log messages, refer to
        DataTypeModel.__init__.log_data
    :param track_info: function - to display processing info
    :return file: file object - to close or continue to read
    """
    byteorder = trace.stats.mseed['byteorder']
    h = trace.stats
    logText = "\n\nSTATE OF HEALTH: "
    logText += ("From:%s  To:%s\n" % (h.starttime, h.endtime))
    textFromData = trace.data.tobytes().decode()
    if textFromData != '':
        logText += textFromData
    else:
        recLen = h.mseed['record_length']
        if file is None:
            file = open(path2file, 'rb')
        databytes = file.read(recLen)
        followingBlktsTotal = unpack('%s%s' % (byteorder, 'B'),
                                     databytes[39:40])[0]
        if followingBlktsTotal > 1:
            nextBlktByteNo = 48 + 8  # header + blkt1000(SEED info)
        else:
            nextBlktByteNo = 0
        while nextBlktByteNo != 0:
            try:
                nextBlktByteNo, info = readNextBlkt(
                    nextBlktByteNo, databytes, byteorder)
                logText += info
            except ReadBlocketteError as e:
                track_info(f"{sta_id} - {chan_id}: {e.msg}", LogType.ERROR)

    if sta_id not in log_data:
        log_data[sta_id] = {}
    if h.channel not in log_data[sta_id]:
        log_data[sta_id][h.channel] = []
    log_data[sta_id][h.channel].append(logText)
    return file


def readText(path2file: Path, fileName: str, textLogs: List) -> bool:
    """
    Read text file and add to logData under channel TEXT. Raise exception
        if the file isn't a text file
    :param path2file: str - absolute path to text file
    :param fileName: str - name of text file
    :param textLogs: holder to keep log string, refer to
        DataTypeModel.__init__.log_data['TEXT']
    """
    with open(path2file, 'r') as file:
        try:
            content = file.read()
        except UnicodeDecodeError:
            raise Exception("Can't process file: %s" % fileName, LogType.ERROR)

        logText = "\n\n** STATE OF HEALTH: %s\n" % fileName
        logText += content
        textLogs.append(logText)

        return True


def saveData2File(tmp_dir: str, tm_data: str, sta_id: str, chanid: str,
                  tr: np.ndarray, tr_idx: int, tr_size: int) -> Path:
    """
    Using np.memmap save time/data to file then free memory for processing
    :param tmp_dir: str - the temporary dir to save file in
    :param tm_data: str - "times"/"data"
    :param sta_id: str - station ID
    :param chanid: str - channel ID
    :param tr: numpy array - of trace time or data
    :param tr_idx: int - trace index
    :param tr_size: int - trace size
    :return memFileName: str - name of memmap file
    """
    memFileName = Path(tmp_dir).joinpath(
        f"{sta_id}-{chanid}-{tm_data}-{tr_idx}")
    memFile = np.memmap(memFileName, dtype='int64', mode='w+',
                        shape=tr_size)
    memFile[:] = tr[:]
    del tr
    del memFile
    return memFileName


def checkChan(chanID: str, reqSOHChans: List[str], reqWFChans: List[str]
              ) -> Union[str, bool]:
    """
    Check if chanID is a requested channel.
    :param chanID: str - channel ID
    :param reqSOHChans: list of str - requested SOH channels
    :param reqWFChans: list of str - requested waveform channels
    :return: str/bool -
        'WF' if chanID is a requested waveform channel,
        'SOH' if chanID is a requested SOH channel,
        False otherwise.
    """
    ret = checkWFChan(chanID, reqWFChans)
    if ret[0] == 'WF':
        if ret[1]:
            return "WF"
        else:
            return False
    if checkSOHChan(chanID, reqSOHChans):
        return "SOH"
    return False


def checkSOHChan(chanID: str, reqSOHChans: List[str]) -> bool:
    """
    Check if chan_id is a requested SOH channel.
    Mass position is always included.
    This function is used for mseed only so mass position is 'VM'.
    If there is no reqSOHChans, it means all SOH channels are requested
    :param chanID: str - channel ID
    :param reqSOHChans: list of str - requested SOH channels
    :return: bool - True if chan_id is a requested SOH channel. False otherwise
    """
    if reqSOHChans == []:
        return True
    if chanID in reqSOHChans:
        return True
    if 'EX?' in reqSOHChans and chanID.startswith('EX'):
        if chanID[2] in ['1', '2', '3']:
            return True
    if chanID.startswith('VM'):
        if chanID[2] in ['0', '1', '2', '3', '4', '5', '6']:
            return True
    return False


def checkWFChan(chanID: str, reqWFChans: List[str]) -> Tuple[str, bool]:
    """
    Check if chanID is a waveform channel and is requested by user
    :param chanID: str - channel ID
    :param reqWFChans: list of str - requested waveform channels
    :return wf: str - '' if chan_id is not a waveform channel.
                      'WF' if chan_id is a waveform channel.
    :return hasChan: bool - True if chan_id is a requested waveform channel.
    """
    wf = ''
    hasChan = False
    if dbConf['seisRE'].match(chanID):
        wf = 'WF'
    if reqWFChans == ['*']:
        hasChan = True
    if chanID in reqWFChans:
        hasChan = True
    return wf, hasChan


def sortData(dataDict: Dict) -> None:
    """
    Sort data in 'tracesInfo' in 'startTmEpoch' order
    :param dataDict: DataTypeModel.__init__.waveformData
    """
    for staID in dataDict:
        for chanID in dataDict[staID]['readData']:
            tracesInfo = dataDict[staID]['readData'][chanID]['tracesInfo']
            tracesInfo = sorted(
                tracesInfo, key=lambda i: i['startTmEpoch'])


def squash_gaps(gaps: List[List[float]]) -> List[List[float]]:
    """
    Compress gaps from different channels that have time range related to
    each other to the ones with outside boundary (min start, max end).
    :param gaps: [[float, float],], [[float, float],] -
        list of gaps of multiple channels: [[start, end],], [[start, end],]
    :return: squashed_gaps: [[float, float],] - all related gaps are squashed
        extending to min start and max end [[min start, max end],]
    """
    squashed_gaps = []
    sgap_indexes = []
    for idx, g in enumerate(gaps):
        if idx in sgap_indexes:
            continue
        squashed_gaps.append(g)
        sgap_indexes.append(idx)
        for idx_, g_ in enumerate(gaps):
            if idx_ in sgap_indexes:
                continue
            if ((g[0] <= g_[0] <= g[1]) or (g[0] <= g_[1] <= g[1])
                    or (g_[0] <= g[0] <= g_[1]) or (g_[0] <= g[1] <= g_[1])):
                sgap_indexes.append(idx_)
                squashed_gaps[-1][0] = min(g[0], g_[0])
                squashed_gaps[-1][1] = max(g[1], g_[1])
            else:
                break
    return squashed_gaps


def downsample(times: np.ndarray, data: np.ndarray,
               log_indexes: Optional[np.ndarray] = None, rq_points: int = 0
               ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Reduce sample rate of times and data so that times and data return has
        the size around the rq_points.
    Since the functions used for downsampling (chunk_minmax()/constant_rate)
        are very slow, the values of data from mean to CUT_FROM_MEAN_FACTOR
        will be removed first. If the size not meet the rq_points, then
        continue to downsample.
    :param times: numpy array - of a waveform channel's times
    :param data: numpy array - of a waveform channel's data
    :param log_indexes: numpy array - of a waveform channel's soh message line
        index
    :param rq_points: int - requested size to return.
    :return np.array, np.array,(np.array) - new times and new data (and new
        log_indexes) with the requested size
    """
    # create a dummy array for log_indexes. However this way may slow down
    # the performance of waveform downsample because waveform channel are large
    # and have no log_indexes.
    if log_indexes is None:
        log_indexes = np.empty(times.size)
    if times.size <= rq_points:
        return times, data, log_indexes
    dataMax = max(abs(data.max()), abs(data.min()))
    dataMean = abs(data.mean())
    indexes = np.where(
        abs(data - data.mean()) >
        (dataMax - dataMean) * const.CUT_FROM_MEAN_FACTOR)
    times = times[indexes]
    data = data[indexes]
    log_indexes = log_indexes[indexes]

    if times.size <= rq_points:
        return times, data, log_indexes

    return chunk_minmax(times, data, log_indexes, rq_points)


def constant_rate(times: np.ndarray, data: np.ndarray, rq_points: int
                  ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Take sample with constant_rate regardless of the value of the data
    :param times: numpy array of a waveform channel's times
    :param data: numpy array of a waveform channel's data
    :param rq_points: requested size to return.
    :return times, data: new times and new data with the requested size
    """
    if times.size <= rq_points:
        return times, data
    rate = int(times.size/rq_points)
    if rate == 1:
        return times, data

    indexes = np.arange(0, times.size, rate)
    times = times[indexes]
    data = data[indexes]

    return times, data


def chunk_minmax(times: np.ndarray, data: np.ndarray,
                 log_indexes: Optional[np.ndarray],
                 rq_points: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Split data into different chunks, take the min, max of each chunk to add
        to the data return
    :param times: numpy array - of a channel's times
    :param data: numpy array - of a channel's data
    :param log_indexes: numpy array - of a channel's log_indexes
    :param rq_points: int - requested size to return.
    :return times, data: np.array, np.array - new times and new data with the
        requested size
    """
    x, y, z = times, data, log_indexes
    final_points = 0
    if x.size <= rq_points:
        final_points += x.size
        return x, y, log_indexes

    if rq_points < 2:
        return np.empty((1, 0)), np.empty((1, 0)), np.empty((1, 0))

    # Since grabbing the min and max from each
    # chunk, need to div the requested number of points
    # by 2.
    size = rq_points // 2
    cs = math.ceil(times.size / size)

    if cs * size > times.size:
        cs -= 1
        # Length of the trace is not divisible by the number of requested
        # points. So split into an array that is divisible by the requested
        # size, and an array that contains the excess. Downsample both,
        # and combine. This case gives slightly more samples than
        # the requested sample size, but not by much.
        x0 = times[:cs * size]
        y0 = data[:cs * size]
        z0 = log_indexes[:cs * size]

        x1 = times[cs * size:]
        y1 = data[cs * size:]
        z1 = data[cs * size:]

        dx0, dy0, dz0 = downsample(x0, y0, z0, rq_points=rq_points)

        # right-most subarray is always smaller than
        # the initially requested number of points.
        dx1, dy1, dz1 = downsample(x1, y1, z1, rq_points=cs)

        dx = np.zeros(dx0.size + dx1.size)
        dy = np.zeros(dy0.size + dy1.size)
        dz = np.zeros(dz0.size + dz1.size)

        dx[:dx0.size] = dx0
        dy[:dy0.size] = dy0
        dz[:dz0.size] = dz0

        dx[dx0.size:] = dx1
        dy[dy0.size:] = dy1
        dz[dz0.size:] = dz1

        return dx, dy, dz

    x = x.reshape(size, cs)
    y = y.reshape(size, cs)
    z = z.reshape(size, cs)

    imin = np.argmin(y, axis=1)
    imax = np.argmax(y, axis=1)

    rows = np.arange(size)

    mask = np.zeros(shape=(size, cs), dtype=bool)
    mask[rows, imin] = True
    mask[rows, imax] = True

    dx = x[mask]
    dy = y[mask]
    dz = z[mask]
    return dx, dy, dz


def trim_downsample_SOHChan(chan: Dict, startTm: float, endTm: float,
                            firsttime: bool) -> None:
    """
    When channel is zoomed in, times and data will be trimmed off to exclude
        the non-included time. Then they will be downsampled to get to the
        expected const.CHAN_SIZE_LIMIT, and save to chan[times], chan[data]
        for plotting.

    :param chan: dict - channel info from soh_data, refer
        DataType_Model.__init_.soh_data[key][chan_id]
    :param startTm: float - start time of zoomed section
    :param endTm: float - end time of zoomed section
    :param firsttime: bool True for original size when channel is not zoomed in
    """
    # zoom into the given time
    tr = chan['orgTrace']
    indexes = np.where((startTm <= tr['times']) & (tr['times'] <= endTm))
    if 'logIdx' in tr.keys():
        chan['times'], chan['data'], chan['logIdx'] = downsample(
            tr['times'][indexes], tr['data'][indexes], tr['logIdx'][indexes],
            rq_points=const.CHAN_SIZE_LIMIT)
    else:
        chan['times'], chan['data'], _ = downsample(
            tr['times'][indexes], tr['data'][indexes],
            rq_points=const.CHAN_SIZE_LIMIT)


def trim_waveform_data(wf_channel_data: Dict, start_time: float,
                       end_time: float) -> List[Dict]:
    """
    Trim off waveform traces whose times does not intersect the closed interval
    [start_time, end_time].

    :param wf_channel_data: dict - a dictionary that contains the waveform
        information of a channel. For its structure, refer to
        DataTypeModel.waveformData[key]['readData'][chan_id].
    :param start_time: the start of the time range to trim waveform data to.
    :param end_time: the end of the time range to trim waveform data to.
    :return: a list with all the waveform traces that were not trimmed.
    """
    if ((start_time > wf_channel_data['tracesInfo'][-1]['endTmEpoch']) or
            (end_time < wf_channel_data['tracesInfo'][0]['startTmEpoch'])):
        return []

    good_start_indices = [index
                          for index, tr
                          in enumerate(wf_channel_data['tracesInfo'])
                          if tr['startTmEpoch'] > start_time]
    if good_start_indices:
        start_idx = good_start_indices[0]
        if start_idx > 0:
            start_idx -= 1  # start_time in middle of trace
    else:
        start_idx = 0

    good_end_indices = [idx
                        for idx, tr
                        in enumerate(wf_channel_data['tracesInfo'])
                        if tr['endTmEpoch'] <= end_time]
    if good_end_indices:
        end_idx = good_end_indices[-1]
        if end_idx < len(wf_channel_data['tracesInfo']) - 1:
            end_idx += 1  # end_time in middle of trace
    else:
        end_idx = 0
    end_idx += 1  # a[x:y+1] = [a[x], ...a[y]

    trimmed_traces_list = wf_channel_data['tracesInfo'][start_idx:end_idx]

    return trimmed_traces_list


def downsample_waveform_data(trimmed_traces_list: List[Dict],
                             start_time: float, end_time: float
                             ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Downsample trimmed waveform data. Works on one trace at a time instead
    of combining the traces before downsampling.

    :param trimmed_traces_list: a list of traces that have been trimmed to
        include only those that intersect the closed interval
        [start_time, end_time]
    :param start_time: the start of the time range
    :param end_time: the end of the time range
    :return: list of downsampled times array and list of downsampled data array
    """
    # Calculate the number of requested_points
    total_size = sum([tr['size'] for tr in trimmed_traces_list])
    requested_points = 0
    if total_size > const.CHAN_SIZE_LIMIT:
        requested_points = int(
            const.CHAN_SIZE_LIMIT / len(trimmed_traces_list)
        )

    # Downsample the data
    downsampled_times_list = []
    downsampled_data_list = []
    for trIdx, tr in enumerate(trimmed_traces_list):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        indexes = np.where((start_time <= times) & (times <= end_time))
        times = times[indexes]
        data = data[indexes]
        if requested_points != 0:
            times, data, _ = downsample(times, data,
                                        rq_points=requested_points)
        downsampled_times_list.append(times)
        downsampled_data_list.append(data)

    return downsampled_times_list, downsampled_data_list


def trim_downsample_WFChan(chan: Dict, startTm: float, endTm: float,
                           firsttime: bool) -> Optional[bool]:
    """
    When channel is zoomed in, times and data will be trimmed off to exclude
        the non-included time.
    Different with soh_data where times and data are each in one np.array,
        in waveform_data, times and data are each kept in a list of np.memmap
        files along with startTmEpoch and endTmEpoch.
    Data in np.memmap files will be trimmed again if there is still time
        outside the zoom time range then downsampled and combined to get the
        times and data with size closed to the expected const.CHAN_SIZE_LIMIT,
        and save to chan[times], chan[data] for plotting.
    chan['fulldata']: is a flag set at the firsttime when channel isn't
        zoomed in and the data is small enough so it doesn't need to go through
        the process when zoomed in.

    :param chan: dict - channel info from soh_data, refer
        DataType_Model.__init_.soh_data[key][chan_id]
    :param startTm: float - start time of zoomed section
    :param endTm: float - end time of zoomed section
    :param firsttime: bool True for original size when channel is not zoomed in
    """
    if 'fulldata' in chan:
        # data is small, already has full in the first trim
        return

    zTracesInfo = trim_waveform_data(chan, startTm, endTm)
    if not zTracesInfo:
        return False

    totalSize = sum([tr['size'] for tr in zTracesInfo])
    if not firsttime and totalSize > const.RECAL_SIZE_LIMIT:
        # size too big, recalc only when it's small enough
        return
    if totalSize <= const.CHAN_SIZE_LIMIT and firsttime:
        chan['fulldata'] = True

    try:
        del chan['times']
        del chan['data']
    except Exception:
        pass

    downsampled_times, downsampled_data = downsample_waveform_data(
        zTracesInfo, startTm, endTm
    )

    chan['times'] = np.hstack(downsampled_times)
    chan['data'] = np.hstack(downsampled_data)


def get_eachDay5MinList(startTm: float, endTm: float) -> np.ndarray:
    """
    Get the list of all five minute for every day start from the day of startTm
        and end at the day of endTm.
    :param startTm: float - start time
    :param endTm: float - end time
    :return every_day_5_min_list: [[288 of floats], ] - the list of all start
        of five minutes for every day in which each day has 288 of 5 minutes.
    """
    exactDayTm = (startTm // const.SEC_DAY) * const.SEC_DAY
    exactDayTmList = []

    if startTm < exactDayTm:
        exactDayTmList = [exactDayTm - const.SEC_DAY]

    while exactDayTm < endTm:
        exactDayTmList.append(exactDayTm)
        exactDayTm += const.SEC_DAY

    # list of start/end 5m in each day: every_day_5_min_list
    for idx, startDayTm in enumerate(exactDayTmList):
        aDay5Min = np.arange(startDayTm,
                             startDayTm + const.SEC_DAY,
                             const.SEC_5M)
        if idx == 0:
            every_day_5_min_list = np.array([aDay5Min])
        else:
            every_day_5_min_list = np.vstack((every_day_5_min_list, aDay5Min))
    return every_day_5_min_list


def get_trimTPSData(chan: Dict, startTm: float, endTm: float,
                    every_day_5_min_list: List[List[float]]) -> Optional[bool]:
    """
    Different with soh_data where times and data are each in one np.array,
        in waveform_data, times and data are each kept in a list of np.memmap
        files along with startTmEpoch and endTmEpoch.
    chan['startIdx'] and chan['endIdx'] will be identify to exclude np.memmap
        files that aren't in the zoom time range (startTm, endTm).
    Data in np.memmap will be trimmed according to times then time-power-square
        value for each 5 minutes will be calculated and saved in
        chan['tps-data']: np.mean(np.square(5m data))

    :param chan: dict - chan dictionary, refer to
        DataTypeModel.__init__.waveformData[key]['readData'][chan_id]
    :param startTm: float - start time of zoomed section
    :param endTm: float - end time of zoomed section
    :param every_day_5_min_list: [[288 of floats], ] - the list of all start
        of five minutes for every day in which each day has 288 of 5 minutes.
    """
    # preset all 0 for all 5 minutes for each day
    tpsData = np.zeros((len(every_day_5_min_list), const.NO_5M_DAY))

    # zoom in to the given range
    chan['startIdx'] = 0
    chan['endIdx'] = len(chan['tracesInfo'])

    if ((startTm > chan['tracesInfo'][-1]['endTmEpoch']) or
            (endTm < chan['tracesInfo'][0]['startTmEpoch'])):
        return False

    indexes = [index for index, tr in enumerate(chan['tracesInfo'])
               if tr['startTmEpoch'] > startTm]
    if indexes != []:
        chan['startIdx'] = indexes[0]
        if chan['startIdx'] > 0:
            chan['startIdx'] -= 1  # startTm in middle of trace
    else:
        chan['startIdx'] = 0

    indexes = [idx for (idx, tr) in enumerate(chan['tracesInfo'])
               if tr['endTmEpoch'] <= endTm]
    if indexes != []:
        chan['endIdx'] = indexes[-1]
        if chan['endIdx'] < len(chan['tracesInfo']) - 1:
            chan['endIdx'] += 1  # endTm in middle of trace
    else:
        chan['endIdx'] = 0
    chan['endIdx'] += 1  # a[x:y+1] = [a[x], ...a[y]

    zTracesInfo = chan['tracesInfo'][chan['startIdx']:chan['endIdx']]

    spr = chan['samplerate']
    chan['tps_data'] = []

    startTPSTm = 0
    accDataList = []

    for trIdx, tr in enumerate(zTracesInfo):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        startIndex = 0
        if trIdx == 0:
            # get index of times with closet value to startTm
            startIndex = np.abs(times - startTm).argmin()
            startTPSTm = times[startIndex]

        # identify index in case of overlaps or gaps
        index = np.where((every_day_5_min_list <= times[startIndex]) &
                         (every_day_5_min_list + const.SEC_5M >
                          times[startIndex]))
        currRow = index[0][0]
        currCol = index[1][0]
        nextTPSTm = startTPSTm + const.SEC_5M
        while endTm >= nextTPSTm:
            nextIndex = int(startIndex + spr * const.SEC_5M)
            if nextIndex >= tr['size']:
                accDataList.append(data[startIndex:tr['size']])
                break
            else:
                accDataList.append(np.square(data[startIndex:nextIndex]))
                accData = np.hstack(accDataList)
                if accData.size == 0:
                    tpsData[currRow, currCol] = 0
                else:
                    tpsData[currRow, currCol] = np.mean(accData)

                startIndex = nextIndex
                currCol += 1
                accDataList = []
                if currCol == const.NO_5M_DAY:
                    currCol = 0
                    currRow += 1
                    nextTPSTm += const.SEC_5M
    chan['tps_data'] = tpsData


def findTPSTm(given_tm: float, each_day_5_min_list: List[List[float]]
              ) -> Tuple[float, float]:
    """
    Find the position of the given time (given_tm) in time-power-squared plot
    :param given_tm: float - given time
    :param each_day_5_min_list: [[288 of floats], ] - the list of all start
        of five minutes for every day in which each day has 288 of 5 minutes.
    :return x_idx: int - index of time in the each_day_5_min_list
    :return y_idx: int - index of day plotted
                   (look in TimePowerSquaredWidget.getZoomData())
    """
    x_idx = None
    y_idx = None
    for day_idx, day in enumerate(each_day_5_min_list):
        for tm_idx, tm in enumerate(day):
            if tm > given_tm:
                y_idx = - day_idx
                x_idx = tm_idx - 1
                if x_idx < 0:
                    y_idx = -(day_idx - 1)
                    x_idx = len(day) - 1
                break
        if x_idx is not None:
            break
    return x_idx, y_idx


if __name__ == '__main__':
    print(checkChan('HH1', [], []))
