"""
MSeed object to hold and process MSeed data
"""
import functools
import operator
import os
from pathlib import Path
from typing import Dict, Tuple, List, Set

from obspy.core import Stream

from sohstationviewer.conf import constants
from sohstationviewer.controller.util import validateFile
from sohstationviewer.model.data_type_model import DataTypeModel, ThreadStopped
from sohstationviewer.model.gps_point import GPSPoint
from sohstationviewer.model.handling_data import (
    readWaveformMSeed, squash_gaps, checkWFChan, sortData, readSOHTrace,
)
from sohstationviewer.model.mseed.from_mseedpeek.mseed_header import readHdrs
from sohstationviewer.view.util.enums import LogType


class MSeed(DataTypeModel):
    """
    read and process mseed file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # channels: available channels got from files among
        # requested channels
        self.channels: Set[str] = set()

        # nets: available nets got from files
        self.nets: Set[str] = set()

        # stats: available stats got from files
        self.stats: Set[str] = set()

        """
        A file can have more than one experiment which only one of them should
        be correct as selected by user (handling_data.py - read_soh_mseed())
        netsProbInFile is the dictionary with key is the tuple of all nets
        in a file, value is the selected net selected by user.
        """
        self.netsProbInFile: Dict[Tuple[str, ...], str] = {}

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.read_soh_and_index_waveform(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selectedKey = self.selectStaID()

        if self.selectedKey is None:
            raise ThreadStopped()
        if len(self.reqWFChans) != 0:
            self.readWFFiles(self.selectedKey)

        self.get_gps_data_q330()

    def read_soh_and_index_waveform(self, folder: str):
        """
            + read waveform data for filename associate with time range
            + read soh_stream for unmerged traces
            + merge soh_stream into SOH data and mass position data
            + squash gaps from different soh channels to one
        :param folder: absolute path to data set folder
        """
        self.waveformData, soh_stream = self.index_waveform(folder)
        self.SOHData, self.massPosData, self.gaps = self.merge_soh_streams(
            soh_stream)

    def index_waveform(self, folder: str
                       ) -> Tuple[Dict, Dict[str, Dict[str, Stream]]]:
        """
        :param folder: absolute path to data set folder
        :return waveformData: a dict for waveform data including
        {sta_id: {'filesInfo': {chan_id: [{filepath,time range,trace's info}]}
                 'readData': { chan_id: {samplerate
                         and preset traces_info which is a list of {info and
                         np.memmap's file path of trace data and time}
                  }
        }}
            (use np.memmap because data are too large to be merged and
                processed)
        :return soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        """
        soh_streams = {}
        waveform_data = {}
        net_stat_prop = {}
        stat_prop = {}
        chan_prop = {}
        count = 0

        for path, sub_dirs, files in os.walk(folder):
            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()

                path2file = Path(path).joinpath(file_name)
                if not validateFile(path2file, file_name):
                    continue
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} file headers/ SOH files', LogType.INFO)

                ret = readHdrs(
                    path2file, file_name, soh_streams, self.logData,
                    self.reqSOHChans, self.reqWFChans,
                    self.netsProbInFile, self.trackInfo)
                if ret is None:
                    continue

                self.nets.update(ret['nets'])
                self.stats.add(ret['stats'][0])

                if len(ret['stats']) > 1:
                    k = tuple(ret['stats'])
                    if k not in stat_prop:
                        stat_prop[k] = []
                    stat_prop[k].append(file_name)
                else:
                    if len(ret['netStats']) > 1:
                        k = tuple(ret['netStats'])
                        if k not in net_stat_prop:
                            net_stat_prop[k] = []
                        net_stat_prop[k].append(file_name)
                if len(ret['chanIDs']) > 1:
                    k = tuple(ret['chanIDs'])
                    if k not in chan_prop:
                        chan_prop[k] = []
                    chan_prop[k].append(file_name)
                self.channels.update(ret['chanIDs'])
                if 'read' not in ret:
                    # not waveform
                    continue

                # TODO currently for waveform not read the file
                #  with 2 stations, 2 channels
                # correctly. Use the first one in the file only
                # When have time will dig more to this
                sta_id = ret['stats'][0]
                chan_id = ret['chanIDs'][0]
                if sta_id not in waveform_data:
                    waveform_data[sta_id] = {"filesInfo": {},
                                             "readData": {}}
                    self.dataTime[sta_id] = [constants.HIGHEST_INT, 0]
                if chan_id not in waveform_data[sta_id]["filesInfo"]:
                    waveform_data[sta_id]["filesInfo"][chan_id] = []
                    waveform_data[sta_id]["readData"][chan_id] = {
                        "samplerate": ret['spr'],
                        "tracesInfo": []}
                waveform_data[sta_id]['filesInfo'][chan_id].append(ret)

        if len(stat_prop) > 0:
            errmsg = (f"More than one stations in a file: {stat_prop}. "
                      f"Will use the first one.")
            self.trackInfo(errmsg, LogType.ERROR)
        if len(net_stat_prop) > 0:
            errmsg = "More than one netIDs in a file: %s" % net_stat_prop
            self.trackInfo(errmsg, LogType.WARNING)
        if len(chan_prop) > 0:
            errmsg = (f"More than one channels in a file: {chan_prop} "
                      f"\nThis is a CRITICAL ERROR.")
            self.trackInfo(errmsg, LogType.ERROR)
        return waveform_data, soh_streams

    def merge_soh_streams(self, soh_streams: Dict[str, Dict[str, Stream]]
                          ) -> Tuple[dict, dict, dict]:
        """
        :param soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        if there is more than one network for a station, user can either choose
        data for one network for that station, or combine all networks related
        to that station into one data set (by changing network name to the
        selected network then merge again)
        :return masspos_data: merged data for chan_id with 'VM' prefixed
            (structure in DataTypeModel.__init__.massPosData)
        :return soh_data: merged data for other soh chan_ids
            (structure in DataTypeModel.__init__SOHData)
        :return gaps: start and end of gaps from all merged soh streams
            are squashed to the largest gaps
            (structure in DataTypeModel.__init__gaps)
        """

        soh_data = {}
        masspos_data = {}
        gaps = {}
        all_gaps: List[List[float]] = []
        for sta_id in soh_streams:
            self.stats.add(sta_id)
            self.dataTime[sta_id] = [constants.HIGHEST_INT, 0]
            soh_data[sta_id] = {}
            masspos_data[sta_id] = {}
            for chan_id in soh_streams[sta_id]:
                self.channels.add(chan_id)
                stream = soh_streams[sta_id][chan_id]

                stream.merge()

                tr = stream[0]
                if len(stream) > 1:
                    nets = [tr.stats['network'].strip() for tr in stream]
                    nets += [f"Combine to {n}" for n in nets]
                    msg = (f"There are more than one net for sta {sta_id}.\n"
                           "Please select one or combine all to one.")
                    self.pause_signal.emit(msg, nets)
                    self.pause()
                    sel_net = nets[self.pause_response]
                    if "Combine" not in sel_net:
                        tr = [tr for tr in stream
                              if tr.stats['network'] == sel_net][0]
                        self.nets.add(sel_net)
                    else:
                        for tr in stream:
                            sel_net = sel_net.replace("Combine to ", "")
                            tr.stats['network'] = sel_net
                            self.nets.add(sel_net)
                        stream.merge()
                        tr = stream[0]

                gaps_in_stream = stream.get_gaps()
                all_gaps += [[g[4].timestamp, g[5].timestamp]
                             for g in gaps_in_stream]
                trace_info = readSOHTrace(tr)
                if chan_id.startswith('VM'):
                    masspos_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                else:
                    soh_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                self.dataTime[sta_id][0] = min(trace_info['startTmEpoch'],
                                               self.dataTime[sta_id][0])
                self.dataTime[sta_id][1] = max(trace_info['endTmEpoch'],
                                               self.dataTime[sta_id][1])

            gaps[sta_id] = squash_gaps(all_gaps)
        return soh_data, masspos_data, gaps

    def selectStaID(self) -> str:
        """
        :return selectedStaID: the selected station id from self.stats
            + If there is only one station id, return it.
            + If there is more than one, show all ids, let user choose one to
                return.
        """

        stats = list(self.stats)
        selectedStaID = stats[0]
        if len(stats) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, stats)
            self.pause()
            selectedStaID = stats[self.pause_response]

        self.trackInfo(f'Select Station {selectedStaID}', LogType.INFO)
        return selectedStaID

    def readWFFiles(self, staID: str) -> None:
        """
        From filesInfo, read all waveform data of requested waveform channels
        for given sta_id, in the selected time (from Main Window) to add
        to waveformData[sat_id]['readData'][chan_id]['tracesInfo'].
        'tracesInfo' includes list of traces' info dict in startTmEpoch order.
        The keys of a trace's info dict:
         {
            'samplerate': sample rate of the data,
            'startTmEpoch': start epoch time of the trace,
            'endTmEpoch': end epoch time of the trace,
            'size': size of data,
            'times_f': np.memmap's file path for time
            'data_f': np.memmap's file path for data
        }
        :param staID: station ID
        """

        count = 0
        for chanID in self.waveformData[staID]['filesInfo']:
            # check chanID
            hasChan = checkWFChan(chanID, self.reqWFChans)
            if not hasChan:
                continue
            tracesInfo = self.waveformData[staID][
                'readData'][chanID]['tracesInfo']

            for fileInfo in self.waveformData[staID]['filesInfo'][chanID]:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped
                # file have been read
                if fileInfo['read']:
                    continue

                # check time
                hasData = False

                if ((self.readStart <= fileInfo['startEpoch'] <= self.readEnd) or       # noqa: E501
                        (self.readStart <= fileInfo['endEpoch'] <= self.readEnd)):      # noqa: E501
                    hasData = True
                if not hasData:
                    continue
                readWaveformMSeed(fileInfo['path2file'], fileInfo['fileName'],
                                  staID, chanID, tracesInfo,
                                  self.dataTime[staID], self.tmpDir)
                fileInfo['read'] = True
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} waveform files', LogType.INFO)

        sortData(self.waveformData)

    def get_gps_data_q330(self):
        """
        Read LOG channel and extract GPS data stored in Q330 data set.
        """
        for station in self.logData:
            if station == 'TEXT':
                continue
            # Q330 log data is composed of a list of string, so we combine them
            # into one big string for ease of processing.
            log_str = functools.reduce(
                operator.iconcat, self.logData[station]['LOG'], ''
            )

            log_lines = [line for line in log_str.splitlines() if line != '']
            for idx, line in enumerate(log_lines):
                if line == "GPS Status":
                    # We are assuming that a GPS status report is 12 lines
                    # long, has a specific format, and is followed by a PLL
                    # status report. This method call checks if these
                    # preconditions hold and raise an error if not.
                    self.check_gps_status_format_q330(log_lines[idx:idx + 13])

                    point = self.extract_gps_point_q330(
                        log_lines[idx:idx + 12]
                    )
                    self.gps_points.append(point)

    @staticmethod
    def extract_gps_point_q330(gps_status_lines: List[str]) -> GPSPoint:
        # Last timemark and fix type are always available, so we have to get
        # them before doing anything else.
        last_timemark = gps_status_lines[11][19:]
        fix_type = gps_status_lines[3][10:]

        # If location data is missing, we set them to 0.
        if gps_status_lines[5] == 'Latitude: ':
            return GPSPoint(last_timemark, fix_type, 0, 0, 0, 0)

        num_sats_used = int(gps_status_lines[8].split(': ')[1])

        # Height is encoded as a float followed by the unit. We
        # don't know how many characters the unit is composed of,
        # so we have to loop through the height string backward
        # until we can detect the end of the height value.
        height_str: str = gps_status_lines[4].split(': ')[1]
        # Start pass the end of the string and look backward one
        # index every iteration so we don't have to add 1 to the
        # final index.
        i = len(height_str)
        current_char = height_str[i - 1]
        while current_char != '.' and not current_char.isnumeric():
            i -= 1
            current_char = height_str[i - 1]
        height = float(height_str[:i])

        # Latitude and longitude are encoded in the format
        # <degree><decimal minute><cardinal direction>. For
        # latitude, <degree> has two characters, while for longitude, <degree>
        # has three.
        # To make the GPS points easier to plot, we convert the latitude and
        # longitude to decimal degree.
        raw_latitude = gps_status_lines[5].split(': ')[1]
        lat_degree = int(raw_latitude[:2])
        lat_minute = float(raw_latitude[2:-1]) / 60
        latitude = lat_degree + lat_minute
        if raw_latitude[-1].lower() == 's':
            latitude = -latitude

        raw_longitude = gps_status_lines[6].split(': ')[1]
        long_degree = int(raw_longitude[:3])
        long_minute = float(raw_longitude[3:-1]) / 60
        longitude = long_degree + long_minute
        if raw_longitude[-1].lower() == 'w':
            longitude = -longitude

        gps_point = GPSPoint(last_timemark, fix_type, num_sats_used,
                             latitude, longitude, height)
        return gps_point

    @staticmethod
    def check_gps_status_format_q330(gps_status_lines: List[str]):
        if gps_status_lines[12].lower() != 'pll status':
            raise ValueError(
                'Q330 log data is malformed. '
                'PLL status does not follow GPS status.'
            )
        if 'fix type' not in gps_status_lines[3].lower():
            raise ValueError(
                'Q330 log data is malformed. '
                'Fix type is not at expected position.'
            )
        if 'height' not in gps_status_lines[4].lower():
            raise ValueError(
                'Q330 log data is malformed. '
                'Height is not at expected position.'
            )
        if 'latitude' not in gps_status_lines[5].lower():
            raise ValueError(
                'Q330 log data is malformed. '
                'Latitude is not at expected position.'
            )
        if 'longitude' not in gps_status_lines[6].lower():
            raise ValueError(
                'Q330 log data is malformed. '
                'Longitude is not at expected position.'
            )
        if 'sat. used' not in gps_status_lines[8].lower():
            raise ValueError(
                'Q330 log data is malformed. '
                'Sat. Used is not at expected position.'
            )
        if 'last gps timemark' not in gps_status_lines[11].lower():
            raise ValueError(
                'Q330 log data is malformed. '
                'Last GPS timemark is not at expected position.'
            )
